import numpy as np

from .graph import GraphBase


class LinePlot(GraphBase):
    def __init__(self, *args, filter_size=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter_size = filter_size

    def draw_values(self, axes, indices, values):
        axes.plot(indices, values)

        if self.filter_size is not None:
            axes.plot(indices, filter_values(values, self.filter_size))


def filter_values(values, filter_size):
    padded_values = np.concatenate(
        [np.full(filter_size // 2, values[0]), np.array(values), np.full(filter_size // 2 - 1, values[-1]),]
    )
    filter_kernel = np.ones((filter_size,)) / filter_size
    return np.convolve(padded_values, filter_kernel, mode="valid")



from collections import deque

import numpy as np


class Logger:
    def __init__(self, name):
        self.name = name

    def __call__(self, value):
        print("{}: {}".format(self.name, value))


class WindowFilterLogger:
    def __init__(self, name, filter_size):
        self.name = name
        self.values = deque(maxlen=filter_size)

    def __call__(self, value):
        self.values.append(value)

        if len(self.values) == self.values.maxlen:
            print("{}: {}".format(self.name, np.mean(self.values)))



import numpy as np

from .graph import GraphBase


class Heatmap(GraphBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.colorbar = None

    def draw_values(self, axes, indices, values):
        pixel_width = 0.5

        left = indices[0] - pixel_width
        right = indices[-1] - pixel_width
        if right == left:
            right = left + 1

        top = pixel_width
        bottom = len(values[0]) + pixel_width

        mappable = axes.imshow(
            np.stack(values).T, interpolation="nearest", aspect="auto", extent=[left, right, top, bottom],
        )

        if self.colorbar is not None:
            self.colorbar.remove()
        self.colorbar = self.container.figure.colorbar(mappable, ax=self.axes)



from collections import deque

import matplotlib.pyplot


class GraphBase:
    def __init__(self, xlabel=None, ylabel=None, update_rate=1, max_size=None, container=None):
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.update_rate = update_rate
        self.container = container

        if self.container is None:
            self.container = global_container_instance()
        self.axes = self.container.add_axes()

        self.count = 0
        self.values = deque(maxlen=max_size)

    def __call__(self, value):
        self.count += 1
        self.values.append(value)
        self.indices = range(self.count - len(self.values), self.count)

        if self.count % self.update_rate == 0:
            self.axes.clear()
            self.axes.set_xlabel(self.xlabel)
            self.axes.set_ylabel(self.ylabel)
            self.draw_values(self.axes, self.indices, self.values)
            self.container.refresh()

    def draw_values(self, axes, indices, values):
        """ to be implemented by concrete type """
        pass


class GraphContainer:
    def __init__(self):
        self.figure = matplotlib.pyplot.figure()
        self.figure.show()

        self.num_rows = 0
        self.num_columns = 1
        self.num_axes = 0

        self.axes_list = []

    def refresh(self):
        self.figure.canvas.draw()

    def add_axes(self):
        self._increment_counts()
        self._update_layout()
        axes = self.figure.add_subplot(self.num_rows, self.num_columns, self.num_axes)
        self.axes_list.append(axes)
        return axes

    def _increment_counts(self):
        self.num_axes += 1
        self.num_rows = self.num_axes // self.num_columns

    def _update_layout(self):
        for i, axes in enumerate(self.axes_list):
            axes.change_geometry(self.num_rows, self.num_columns, i + 1)

        # make some space for labels
        self.figure.tight_layout(rect=(0.05, 0.05, 1, 1))


_global_container_instance = None


def global_container_instance():
    global _global_container_instance

    if _global_container_instance is None:
        _global_container_instance = GraphContainer()

    return _global_container_instance



__recorders = {}


def set_recorder(**kwargs):
    global __recorders
    for name, recorder in kwargs.items():
        __recorders[name] = recorder


def record(**kwargs):
    global __recorders
    for key, value in kwargs.items():
        if key in __recorders.keys():
            __recorders[key](value)
